/** 
 * \file kem.cpp
 * \brief Implementation of api.h
 */

#include "hash.h"
#include "rng.h"
#include "api.h"
#include "parameters.h"
#include "qcrs_recoverer.h"
#include "parsing.h"
#include "ffi_field.h"
#include "ffi_vec.h"
#include "ffi_elt.h"


/** 
 * \fn int crypto_kem_keypair(unsigned char* pk, unsigned char* sk)
 * \brief Keygen of the Loong2 scheme
 *
 * The public key is composed of the the seed used to generate the matrix <b>H</b> and matrix Q.
 * The secret key is composed of the seed used to generate the matrices <b>X</b> and <b>Y</b>.
 *
 * \param[out] pk String containing the public key
 * \param[out] sk String containing the secret key
 * \return 0 if keygen is sucessfull
 */
int crypto_kem_keypair(unsigned char* pk, unsigned char* sk) {
  #ifdef VERBOSE
    printf("\n\n\n### KEYGEN ###");
  #endif

  // Create seed expanders for public key and secret key
  unsigned char sk_seed[SEEDEXPANDER_SEED_BYTES];
  randombytes(sk_seed, SEEDEXPANDER_SEED_BYTES);
  AES_XOF_struct* sk_seedexpander = (AES_XOF_struct*) malloc(sizeof(AES_XOF_struct));
  seedexpander_init(sk_seedexpander, sk_seed, sk_seed + 32, SEEDEXPANDER_MAX_LENGTH);

  unsigned char pk_seed[SEEDEXPANDER_SEED_BYTES];
  randombytes(pk_seed, SEEDEXPANDER_SEED_BYTES);
  AES_XOF_struct* pk_seedexpander = (AES_XOF_struct*) malloc(sizeof(AES_XOF_struct));
  seedexpander_init(pk_seedexpander, pk_seed, pk_seed + 32, SEEDEXPANDER_MAX_LENGTH);

  ffi_field_init();

  // Compute secret key X, Y
  ffi_vec F, X[PARAM_N], Y[PARAM_N];
  ffi_vec_set_random_full_rank_with_one_using_seedexpander(F, PARAM_W, sk_seedexpander);
  for(int i = 0; i < PARAM_N; i++)
  {
	ffi_vec_set_random_from_support_using_seedexpander(X[i], PARAM_N, F, PARAM_W, sk_seedexpander);
	ffi_vec_set_random_from_support_using_seedexpander(Y[i], PARAM_N, F, PARAM_W, sk_seedexpander);
  }
  // Compute public key
  ffi_vec H[PARAM_N], Q[PARAM_N];
  for(int i = 0; i < PARAM_N; i++)
  {
	  H[i].SetLength(PARAM_N);
	  Q[i].SetLength(PARAM_N1);
  }
  ffi_elt sum, tmp, tmp1;
  for(int i = 0; i < PARAM_N; i++)
  {
	  ffi_vec_set_random_using_seedexpander(H[i], PARAM_N, pk_seedexpander); 
  }

  for(int i = 0; i < PARAM_N; i++)
  {
	  for(int j = 0; j < PARAM_N1; j++)
	  {
		  ffi_elt_set_zero(tmp);
		  ffi_elt_set_zero(sum);
		  for(int l = 0; l < PARAM_N1; l++)
		  {
			  ffi_elt_mul(tmp1, H[i][l], X[l][j]);
			  ffi_elt_add(sum, tmp, tmp1);
			  ffi_elt_set(tmp, sum);
		  }
		  ffi_elt_add(Q[i][j], tmp, Y[i][j]);
	  }
  }
  ffi_vec_print(F, PARAM_W);printf("\n");
  // Parse keys to string
  loong2_public_key_to_string(pk, Q, pk_seed);
  loong2_secret_key_to_string(sk, sk_seed);

  free(pk_seedexpander);
  free(sk_seedexpander);

  return 0;
}

/** 
 * \fn int crypto_kem_enc(unsigned char* ct, unsigned char* ss, const unsigned char* pk)
 * \brief Encapsulation of the Loong2 scheme
 *
 * \param[out] ct String containing the ciphertext
 * \param[out] ss String containing the shared secret
 * \param[in] pk String containing the public key
 * \return 0 if encapsulation is sucessfull
 */
int crypto_kem_enc(unsigned char* ct, unsigned char* ss, const unsigned char* pk) {
  #ifdef VERBOSE
    printf("\n\n\n\n### ENCAPS ###");
  #endif

  ffi_field_init();

  // Retrieve h and s from public key
  ffi_vec H[PARAM_N], Q[PARAM_N];
  for(int i = 0; i < PARAM_N; i++)
  {
	  H[i].SetLength(PARAM_N);
	  Q[i].SetLength(PARAM_N1);
  }
  loong2_public_key_from_string(H, Q, pk);

  // Generate r1, r2 and er
  ffi_vec support_we, E1[PARAM_N2], E2[PARAM_N2], R[PARAM_N2];
  ffi_vec_set_random_full_rank_using_rng(support_we, PARAM_WE);
  for(int i = 0; i < PARAM_N2; i++)
  {
	  ffi_vec_set_random_from_support_using_rng(E1[i], PARAM_N, support_we, PARAM_WE);
	  ffi_vec_set_random_from_support_using_rng(E2[i], PARAM_N1, support_we, PARAM_WE);
	  ffi_vec_set_random_from_support_using_rng(R[i], PARAM_N, support_we, PARAM_WE);
  }
 
  // Compute C1 = R*H + E1
  ffi_vec C1[PARAM_N2], C2[PARAM_N2];
  ffi_elt sum, tmp, tmp1;
  for(int i = 0; i < PARAM_N2; i++)
  {
	  C1[i].SetLength(PARAM_N);
	  C2[i].SetLength(PARAM_N1);
  }
  for(int i = 0; i < PARAM_N2; i++)
  {
	  for(int j = 0; j < PARAM_N; j++)
	  {
		  ffi_elt_set_zero(tmp);
		  ffi_elt_set_zero(sum);
		  for(int l = 0; l < PARAM_N; l++)
		  {
			  ffi_elt_mul(tmp1, R[i][l], H[l][j]);
			  ffi_elt_add(sum, tmp, tmp1);
			  ffi_elt_set(tmp, sum);
		  }
		  ffi_elt_add(C1[i][j], sum, E1[i][j]);
	  }
  }

  // Compute C2 = R*Q + E2
  for(int i = 0; i < PARAM_N2; i++)
  {
	  for(int j = 0; j < PARAM_N1; j++)
	  {
		  ffi_elt_set_zero(tmp);
		  ffi_elt_set_zero(sum);
		  for(int l = 0; l < PARAM_N; l++)
		  {
			  ffi_elt_mul(tmp1, R[i][l], Q[l][j]);
			  ffi_elt_add(sum, tmp, tmp1);
			  ffi_elt_set(tmp, sum);
		  }
		  ffi_elt_add(C2[i][j], tmp, E2[i][j]);
	  }
  }

  // Compute ciphertext
  loong2_ciphertext_to_string(ct, C1, C2);

  // Compute shared secret
  ffi_vec_echelonize(support_we, PARAM_WE);
  ffi_vec_print(support_we, PARAM_WE);
  unsigned char support[FFI_VEC_WE_BYTES];
  ffi_vec_to_string(support, support_we, PARAM_WE);
  sha512(ss, support, FFI_VEC_WE_BYTES);

  #ifdef VERBOSE
    printf("\n\nh: "); ffi_vec_print(h, PARAM_N);
    printf("\n\ns: "); ffi_vec_print(s, PARAM_N);
    printf("\n\nE: "); ffi_vec_print(E, PARAM_R);
    printf("\n\nr1: "); ffi_vec_print(r1, PARAM_N);
    printf("\n\nr2: "); ffi_vec_print(r2, PARAM_N);
    printf("\n\ner: "); ffi_vec_print(er, PARAM_N);
    printf("\n\nsr: "); ffi_vec_print(sr, PARAM_N);
    printf("\n\nse: "); ffi_vec_print(se, PARAM_N);
    printf("\n\nct: "); for(int i = 0 ; i < CIPHERTEXT_BYTES ; ++i) printf("%02x", ct[i]);
    printf("\n\nss: "); for(int i = 0 ; i < SHARED_SECRET_BYTES ; ++i) printf("%02x", ss[i]);
  #endif

  return 0;
}

/** 
 * \fn int crypto_kem_dec(unsigned char* ss, const unsigned char* ct, const unsigned char* sk)
 * \brief Decapsulation of the Ouroboros-R scheme
 *
 * \param[out] ss String containing the shared secret
 * \param[in] ct String containing the ciphertext
 * \param[in] sk String containing the secret key
 * \return 0 if decapsulation is successfull, -1 otherwise
 */
int crypto_kem_dec(unsigned char* ss, const unsigned char* ct, const unsigned char* sk) {
  #ifdef VERBOSE
    printf("\n\n\n\n### DECAPS ###");
  #endif

  ffi_field_init();

  // Retrieve sr and se from ciphertext
  ffi_vec C1[PARAM_N2], C2[PARAM_N2];
  for(int i = 0; i < PARAM_N2; i++)
  {
	  C1[i].SetLength(PARAM_N);
	  C2[i].SetLength(PARAM_N1);
  }
  loong2_ciphertext_from_string(C1, C2, ct);

  // Retrieve x, y and their support from secret key
  ffi_vec F, X[PARAM_N], Y[PARAM_N];
  F.SetLength(PARAM_W);
  for(int i = 0; i < PARAM_N; i++)
  {
	  X[i].SetLength(PARAM_N1);
	  Y[i].SetLength(PARAM_N1);
  }
  loong2_secret_key_from_string(X, Y, F, sk);

  // Compute C2 - C1*X
  ffi_vec ec[PARAM_N2];
  ffi_elt sum, tmp,tmp1;
  for(int i = 0; i < PARAM_N2; i++)
  {
	  ec[i].SetLength(PARAM_N1);
  }
  for(int i = 0; i < PARAM_N2; i++)
  {
	  for(int j = 0; j < PARAM_N1; j++)
	  {
		  ffi_elt_set_zero(tmp);
		  ffi_elt_set_zero(sum);
		  for(int l = 0; l < PARAM_N; l++)
		  {
			  ffi_elt_mul(tmp1, C1[i][l], X[l][j]);
			  ffi_elt_add(sum, tmp, tmp1);
			  ffi_elt_set(tmp, sum);
		  }
		  ffi_elt_sub(ec[i][j], C2[i][j], tmp);
	  }
  }

  #ifdef VERBOSE
    printf("\n\nsk: "); for(int i = 0 ; i < SECRET_KEY_BYTES ; ++i) printf("%02x", sk[i]);
    printf("\n\nciphertext: "); for(int i = 0 ; i < CIPHERTEXT_BYTES ; ++i) printf("%02x", ct[i]);
    printf("\n\nsr: "); ffi_vec_print(sr, PARAM_N);
    printf("\n\nse: "); ffi_vec_print(se, PARAM_N);
    printf("\n\nF: "); ffi_vec_print(F, PARAM_W);
    printf("\n\nx: "); ffi_vec_print(x, PARAM_N);
    printf("\n\ny: "); ffi_vec_print(y, PARAM_N);
    printf("\n\nec: "); ffi_vec_print(ec, PARAM_N);
  #endif

  // Retrieve E by solving quasi-cyclic rank support recovery problem
  ffi_vec vec_ec;
  vec_ec.SetLength(PARAM_N1*PARAM_N2);
  for(int i = 0; i < PARAM_N2; i++)
  {
	  for(int j = 0; j < PARAM_N1; j++)
	  {
		  vec_ec[j + i*PARAM_N1] = ec[i][j]; 
	  }
  }
  ffi_vec_echelonize(F, PARAM_W);
  ffi_vec_print(F, PARAM_W);
  unsigned int E_dim;
  ffi_vec E;
  qcrs_recoverer(E, E_dim, PARAM_WE, F, PARAM_W, vec_ec, PARAM_N1*PARAM_N2);

  // Compute shared secret 
  if(E_dim != 0) {
    unsigned char support[FFI_VEC_WE_BYTES];
    ffi_vec_to_string(support, E, PARAM_WE);
    sha512(ss, support, FFI_VEC_WE_BYTES);

    #ifdef VERBOSE
      printf("\n\nE: "); ffi_vec_print(E, PARAM_R);
      printf("\n\nss: "); for(int i = 0 ; i < SHARED_SECRET_BYTES ; ++i) printf("%02x", ss[i]);
    #endif

    return 0;
  } else {
    memset(ss, 0, SHARED_SECRET_BYTES);

    #ifdef VERBOSE
      printf("\n\nss: "); for(int i = 0 ; i < SHARED_SECRET_BYTES ; ++i) printf("%02x", ss[i]);
    #endif

    return -1;
  }
}

